import pandas as pd
from typing import Any, Dict, List, Optional
from requests import exceptions, request
from stpstone._config.global_slots import YAML_ANBIMA_DATA_API
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.parsers.dicts import HandlingDicts
from stpstone.utils.parsers.json import JsonFiles
from stpstone.utils.parsers.str import StrHandler


class AnbimaDataGen:

    def __init__(
        self,
        str_client_id: str,
        str_client_secret: str,
        str_env: str = "dev",
        bl_debug: bool = False,
        int_chunk: int = 1000,
        str_host_prd: str = "https://api.anbima.com.br/",
        str_host_dev: str = "https://api-sandbox.anbima.com.br/",
    ):
        """
        Anbima data API

        Args:
            str_client_id (str): string with client id
            str_client_secret (str): string with client secret
            str_env (str): string with environment
            bl_debug (bool): boolean with debug mode
            int_chunk (int): integer with chunk size
            str_host_prd (str): string with production host
            str_host_dev (str): string with development host

        Metadata: https://developers.anbima.com.br/api-portal/pt-br
        """
        self.str_client_id = str_client_id
        self.str_client_secret = str_client_secret
        self.bl_debug = bl_debug
        self.int_chunk = int_chunk
        self.str_host_prd = str_host_prd
        self.str_host = locals()[f"str_host_{str_env.lower()}"]
        self.str_token = self.access_token["access_token"]

    @property
    def access_token(
        self, str_app: str = "oauth/access-token", str_method: str = "POST"
    ) -> List[Dict[str, Any]]:
        str_url = self.str_host_prd + str_app
        base64_credentials = StrHandler().base64_encode(
            self.str_client_id, self.str_client_secret
        )
        dict_headers = {
            "Content-Type": "application/json",
            "Authorization": base64_credentials,
        }
        dict_payload = {"grant_type": "client_credentials"}
        resp_req = request(
            method=str_method,
            url=str_url,
            headers=dict_headers,
            data=JsonFiles().dict_to_json(dict_payload),
        )
        if self.bl_debug == True:
            print("TOKEN: {}".format(resp_req.json()["access_token"]))
        resp_req.raise_for_status()
        return resp_req.json()

    def generic_request(self, str_app: str, str_method: str) -> List[Dict[str, Any]]:
        str_url = self.str_host + str_app
        if self.bl_debug == True:
            print(f"URL: {str_url}")
        dict_headers = {
            "accept": "application/json",
            "client_id": self.str_client_id,
            "access_token": self.str_token,
        }
        resp_req = request(method=str_method, url=str_url, headers=dict_headers)
        resp_req.raise_for_status()
        return resp_req.json()


class AnbimaDataFunds(AnbimaDataGen):
    """
    Metadata: https://developers.anbima.com.br/pt/swagger-de-fundos-v2-rcvm-175/#/Notas%20explicativas/buscarNotasExplicativas
    """

    def funds_raw(
        self,
        int_pg: Optional[int] = None,
        str_app: str = "feed/fundos/v2/fundos?size={}&page={}",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        """
        Retrieve available closed and opened end funds

        Args: -

        Returns:
            Json
        """
        return self.generic_request(
            str_app.format(
                self.int_chunk,
                int_pg,
            ),
            str_method,
        )

    def funds_trt(self, int_pg: int = 0) -> pd.DataFrame:
        """
        Treating available closed and opened end funds

        Args:
            int_pg (int): integer with page number

        Returns:
            pd.DataFrame
        """
        # setting variables
        list_ser = list()
        int_fnd = 0
        # looping within all available fund pages
        while True:
            #   requesting for current fund json, in case status code is different
            #       from 2xx, break the loop and return the dataframe
            try:
                json_funds = self.funds_raw(int_pg)
            except exceptions.HTTPError:
                break
            # looping within content dictionaries
            for dict_cnt in json_funds[YAML_ANBIMA_DATA_API["key_content"]]:
                #   setting variables
                dict_aux = dict()
                int_fnd += 1
                #   looping within keys and values from content
                for (
                    YAML_ANBIMA_DATA_API["key_cnt"],
                    data_cnt,
                ) in dict_cnt.items():
                    if isinstance(data_cnt, str):
                        dict_aux[YAML_ANBIMA_DATA_API["key_cnt"]] = (
                            data_cnt.strip()
                        )
                    elif data_cnt is None:
                        dict_aux[YAML_ANBIMA_DATA_API["key_cnt"]] = data_cnt
                    elif isinstance(data_cnt, list):
                        #   looping within classes
                        for i_cls, dict_cls in enumerate(data_cnt):
                            #   looping within classes and appending to serialized list
                            for (
                                YAML_ANBIMA_DATA_API["key_cls"],
                                data_cls,
                            ) in dict_cls.items():
                                if (
                                    YAML_ANBIMA_DATA_API["key_cls"]
                                    != YAML_ANBIMA_DATA_API["key_name_sbclss"]
                                ) and (data_cls is not None):
                                    dict_aux[
                                        YAML_ANBIMA_DATA_API["key_cls"]
                                    ] = data_cls.strip()
                                elif (
                                    YAML_ANBIMA_DATA_API["key_cls"]
                                    != YAML_ANBIMA_DATA_API["key_name_sbclss"]
                                ) and (data_cls is None):
                                    dict_aux[
                                        YAML_ANBIMA_DATA_API["key_cls"]
                                    ] = data_cls
                                elif (
                                    YAML_ANBIMA_DATA_API["key_cls"]
                                    == YAML_ANBIMA_DATA_API["key_name_sbclss"]
                                ) and (isinstance(data_cls, list)):
                                    #   looping within subclasses and copy auxiliary dicitionary in
                                    #       each iteration, in order to renew the subclass
                                    #       info imported
                                    for dict_sbcls in data_cls:
                                        dict_xpt = dict_aux.copy()
                                        for (
                                            YAML_ANBIMA_DATA_API["key_sbcls"],
                                            data_sbcls,
                                        ) in dict_sbcls.items():
                                            dict_xpt = HandlingDicts().merge_n_dicts(
                                                dict_xpt,
                                                {
                                                    "{}_{}".format(
                                                        YAML_ANBIMA_DATA_API["key_name_sbcls"],
                                                        YAML_ANBIMA_DATA_API["key_sbcls"],
                                                    ): data_sbcls
                                                },
                                                {
                                                    YAML_ANBIMA_DATA_API["col_num_fnd"]: int_fnd + 1,
                                                    YAML_ANBIMA_DATA_API["col_num_class"]: i_cls + 1,
                                                    YAML_ANBIMA_DATA_API["col_num_pg"]: int_pg,
                                                },
                                            )
                                        list_ser.append(dict_xpt)
                                elif (
                                    YAML_ANBIMA_DATA_API["key_cls"]
                                    == YAML_ANBIMA_DATA_API["key_name_sbclss"]
                                ) and (data_cls is None):
                                    list_ser.append(
                                        HandlingDicts().merge_n_dicts(
                                            dict_aux,
                                            {
                                                YAML_ANBIMA_DATA_API["key_cls"]: data_cls
                                            },
                                            {
                                                YAML_ANBIMA_DATA_API["col_num_fnd"]: int_fnd + 1,
                                                YAML_ANBIMA_DATA_API["col_num_class"]: i_cls + 1,
                                                YAML_ANBIMA_DATA_API["col_num_pg"]: int_pg,
                                            },
                                        )
                                    )
                                else:
                                    raise Exception(
                                        "Error of content within class, please revise "
                                        + "pg: {} / key: {} / data: {}".format(
                                            int_pg,
                                            YAML_ANBIMA_DATA_API["key_cls"],
                                            data_cls,
                                        )
                                    )
                    else:
                        raise Exception(
                            f"Error of content data type, please revise the data: {data_cnt}"
                        )
            #   adding iterator
            int_pg += 1
        # appending serialized list to pandas dataframe
        df_funds = pd.DataFrame(list_ser)
        # changing columns types
        for col_dt in [
            YAML_ANBIMA_DATA_API["col_fund_closure_dt"],
            YAML_ANBIMA_DATA_API["col_eff_dt"],
            YAML_ANBIMA_DATA_API["col_incpt_dt"],
            YAML_ANBIMA_DATA_API["col_closure_dt"],
            YAML_ANBIMA_DATA_API["col_sbc_incpt_dt"],
            YAML_ANBIMA_DATA_API["col_sbc_closure_dt"],
            YAML_ANBIMA_DATA_API["col_sbc_eff_dt"],
        ]:
            df_funds[col_dt].fillna(
                YAML_ANBIMA_DATA_API["str_dt_fill_na"], inplace=True
            )
            df_funds[col_dt] = [
                DatesBR().str_date_to_datetime(
                    d, YAML_ANBIMA_DATA_API["str_dt_format"]
                )
                for d in df_funds[col_dt]
            ]
        for col_dt in [
            YAML_ANBIMA_DATA_API["col_update_ts"],
            YAML_ANBIMA_DATA_API["col_sbc_update_dt"],
        ]:
            df_funds[col_dt].fillna(
                YAML_ANBIMA_DATA_API["str_ts_fill_na"], inplace=True
            )
            df_funds[col_dt] = [
                DatesBR().timestamp_to_date(
                    d, format=YAML_ANBIMA_DATA_API["str_dt_format"]
                )
                for d in df_funds[col_dt]
            ]
        df_funds.fillna(YAML_ANBIMA_DATA_API["str_fill_na"], inplace=True)
        df_funds = df_funds.astype(
            {
                YAML_ANBIMA_DATA_API["col_fund_code"]: str,
                YAML_ANBIMA_DATA_API["col_type_id"]: str,
                YAML_ANBIMA_DATA_API["col_fund_id"]: str,
                YAML_ANBIMA_DATA_API["col_comp_name"]: str,
                YAML_ANBIMA_DATA_API["col_trade_name"]: str,
                YAML_ANBIMA_DATA_API["col_fund_type"]: str,
                YAML_ANBIMA_DATA_API["col_class_code"]: str,
                YAML_ANBIMA_DATA_API["col_class_id_type"]: str,
                YAML_ANBIMA_DATA_API["col_class_id"]: str,
                YAML_ANBIMA_DATA_API["col_comp_class"]: str,
                YAML_ANBIMA_DATA_API["col_trd_class"]: str,
                YAML_ANBIMA_DATA_API["col_n1_ctg"]: str,
                YAML_ANBIMA_DATA_API["col_subclasses"]: str,
            }
        )
        # removing duplicates
        df_funds.drop_duplicates(inplace=True)
        # returning dataframe
        return df_funds

    def fund_raw(
        self,
        str_code_fnd: str,
        str_app: str = "feed/fundos/v2/fundos/{}/historico",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        return self.generic_request(str_app.format(str_code_fnd), str_method)

    def fund_trt(self, list_code_fnds: list):
        # setting variables
        dict_dfs = dict()
        # looping within the funds codes
        for str_code_fnd in list_code_fnds:
            #   setting variables
            dict_aux = dict()
            list_ser = list()
            dict_dfs[str_code_fnd] = list()
            #   returning fund info
            json_fnd_info = self.fund_raw(str_code_fnd)
            #   looping within json content
            for key_cnt, data_cnt in json_fnd_info.items():
                #   checking data type, when is a list create the dictionary for the
                #       serialized list to be appended into a pandas df
                if isinstance(data_cnt, str) or data_cnt is None:
                    dict_aux[key_cnt] = data_cnt
                elif isinstance(data_cnt, list):
                    dict_xpt = dict_aux.copy()
                    for dict_data in data_cnt:
                        for key_data, data_data in dict_data.items():
                            #   checking wheter the data instance is string, or list
                            if isinstance(data_data, str) or data_data is None:
                                dict_xpt["{}_{}".format(key_cnt, key_data)] = data_data
                            elif isinstance(data_data, list):
                                for dict_hist in data_data:
                                    dict_xpt_2 = dict_xpt.copy()
                                    for key_hist, data_hist in dict_hist.items():
                                        dict_xpt_2[
                                            "{}_{}_{}".format(
                                                key_cnt, key_data, key_hist
                                            )
                                        ] = data_hist
                                    list_ser.append(dict_xpt_2)
                        #   regarding classes first-order dictionary has a key called
                        #       historico_classe, which is a list (in 2024-11-08), it is treated
                        #       separately in the code, in order to create the serialized list
                        if key_data != "classes":
                            list_ser.append(dict_xpt)
                    df_ = pd.DataFrame(list_ser)
                    #   changing data types within columns
                    for col_ in list(df_.columns):
                        #   date
                        if (
                            StrHandler().match_string_like(col_, "*data_*") == True
                        ) and (len(col_) == 10):
                            df_[col_].fillna(
                                YAML_ANBIMA_DATA_API["str_dt_fill_na"],
                                inplace=True,
                            )
                            df_[col_] = [
                                DatesBR().str_date_to_datetime(
                                    d, YAML_ANBIMA_DATA_API["str_dt_format"]
                                )
                                for d in df_[col_]
                            ]
                        #   timestamp
                        elif (
                            (StrHandler().match_string_like(col_, "*data_*") == True)
                            and (StrHandler().match_string_like(col_, "*T*") == True)
                            and (len(col_) > 10)
                        ):
                            df_[col_].fillna(
                                YAML_ANBIMA_DATA_API["str_ts_fill_na"],
                                inplace=True,
                            )
                            df_[col_] = [
                                DatesBR().timestamp_to_date(
                                    d,
                                    format=YAML_ANBIMA_DATA_API[
                                        "str_dt_format"
                                    ],
                                )
                                for d in df_[col_]
                            ]
                        #   float
                        elif (
                            StrHandler().match_string_like(col_, "*percentual_*")
                            == True
                        ):
                            df_[col_].fillna(
                                YAML_ANBIMA_DATA_API["str_float_fill_na"],
                                inplace=True,
                            )
                            df_[col_] = [float(x) for x in df_[col_]]
                        #   string
                        else:
                            df_[col_].fillna(
                                YAML_ANBIMA_DATA_API["str_fill_na"],
                                inplace=True,
                            )
                            df_[col_] = [str(x).strip() for x in df_[col_]]
                    #   appending to list of dataframe
                    dict_dfs[str_code_fnd].append(df_)
                else:
                    raise Exception(
                        "Type of data within the content of the fund {} ".format(str_code_fnd)
                        + "not found, please check the code. DATA_TYPE: {}".format(type(data_cnt))
                    )
        # returning list of dataframes with fund infos
        return dict_dfs

    def fund_hist(
        self,
        str_code_class: str,
        str_app: str = "feed/fundos/v2/fundos/{}/historico",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        return self.generic_request(str_app.format(str_code_class), str_method)

    def segment_investor(
        self,
        str_code_class: str,
        str_app: str = "feed/fundos/v2/fundos/segmento-investidor/{}/patrimonio-liquido",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        return self.generic_request(str_app.format(str_code_class), str_method)

    def time_series_fund(
        self,
        str_date_inf: str,
        str_date_sup: str,
        str_code_class: str,
        str_app: str = "feed/fundos/v2/fundos/{}/serie-historica",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        dict_payload = {
            "size": self.int_chunk,
            "data-inicio": str_date_inf,
            "data-fim": str_date_sup,
        }
        return self.generic_request(
            str_app.format(str_code_class, self.int_chunk),
            str_method=str_method,
            dict_payload=dict_payload,
        )

    def funds_financials_dt(
        self,
        str_date_update: str,
        str_app: str = "feed/fundos/v2/fundos/serie-historica/lote",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        dict_payload = {"data-atualizacao": str_date_update, "size": self.int_chunk}
        return self.generic_request(
            str_app, str_method=str_method, dict_payload=dict_payload
        )

    def funds_registration_data_dt(
        self,
        str_date_update: str,
        str_app: str = "feed/fundos/v2/fundos/dados-cadastrais/lote",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        dict_payload = {"data-atualizacao": str_date_update, "size": self.int_chunk}
        return self.generic_request(
            str_app, str_method=str_method, dict_payload=dict_payload
        )

    @property
    def institutions(
        self,
        str_app: str = "feed/fundos/v2/fundos/instituicoes",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        dict_payload = {"size": self.int_chunk}
        return self.generic_request(
            str_app, str_method=str_method, dict_payload=dict_payload
        )

    def institution(
        self,
        str_ein: str,
        str_app: str = "feed/fundos/v2/fundos/instituicoes/{}",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        dict_payload = {"size": self.int_chunk}
        return self.generic_request(
            str_app.format(str_ein), str_method=str_method, dict_payload=dict_payload
        )

    def explanatory_notes_fund(
        self,
        str_code_class: str,
        str_app: str = "feed/fundos/v2/fundos/{}/notas-explicativas",
        str_method: str = "GET",
    ) -> List[Dict[str, Any]]:
        return self.generic_request(str_app.format(str_code_class), str_method)



from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class HalfAdder(metaclass=TypeChecker):

    def __init__(self, a: int, b: int) -> None:
        """
        Initialize the Half Adder with two binary inputs.
        Args:
            a (int): First binary input (0 or 1)
            b (int): Second binary input (0 or 1)
        Returns:
            None
        """
        self.a = a
        self.b = b

    @property
    def sum(self) -> int:
        """
        Returns the sum output of the half adder.

        The half adder computes the sum of two binary digits (bits) and produces
        two outputs: the sum and the carry. The sum is calculated using the XOR
        operation:
            - Sum (S) = A XOR B
        Args:
            None
        Returns:
            int: The sum of the two binary inputs.
        """
        return self.a ^ self.b

    @property
    def carry(self) -> int:
        """
        Returns the carry output of the half adder.

        The carry output indicates whether there is an overflow from the addition
        of the two bits. The carry is calculated using the AND operation:
            - Carry (C) = A AND B
        Args:
            None
        Returns:
            int: The carry output of the half adder.
        """
        return self.a & self.b


class FullAdder(metaclass=TypeChecker):

    def __init__(self, a: int, b: int, carry_in: int) -> None:
        """
        Initialize the Full Adder with two binary inputs and a carry-in.
        Args:
            a (int): First binary input (0 or 1)
            b (int): Second binary input (0 or 1)
            carry_in (int): Carry input from the previous stage (0 or 1).
        Returns:
            None
        """
        self.a = a
        self.b = b
        self.carry_in = carry_in

    @property
    def sum(self) -> int:
        """
        Returns the sum output of the full adder.

        The full adder computes the sum of two binary digits (bits) and a carry-in
        bit, producing a sum and a carry-out. The sum is calculated using the XOR
        operation:
            - Sum (S) = A XOR B XOR Carry-In
        Args:
            None
        Returns:
            int: The sum of the two binary inputs and the carry-in
        """
        return (self.a ^ self.b) ^ self.carry_in

    @property
    def carry_out(self) -> int:
        """
        Returns the carry output of the full adder.

        The carry output indicates whether there is an overflow from the addition
        of the two bits and the carry-in. The carry-out is calculated using the
        OR operation on the AND results:
            - Carry-Out (Cout) = (A AND B) OR (Carry-In AND (A XOR B))
        Args:
            None
        Returns:
            int: The carry output of the full adder
        """
        return (self.a & self.b) | (self.carry_in & (self.a ^ self.b))


class EightBitFullAdder(metaclass=TypeChecker):

    def __init__(self, a: int, b: int) -> None:
        """
        Initialize the Eight-Bit Full Adder with two 8-bit binary numbers.
        Args:
            a (int): First 8-bit binary number
            b (int): Second 8-bit binary number
        Returns:
            None
        """
        self.a = a
        self.b = b

    @property
    def add(self) -> tuple:
        """
        Adds two 8-bit numbers and returns the sum and carry out.

        This method iterates through each bit of the two 8-bit numbers, using
        a full adder to compute the sum and carry for each bit position. The
        final result is an 8-bit sum and a carry-out bit.
        Args:
            None
        Returns:
            tuple: A tuple containing the sum as an integer and the carry-out bit.
        """
        carry = 0
        sum_result = 0
        for i in range(8):
            # extract the i-th bit from a and b
            bit_a = (self.a >> i) & 1
            bit_b = (self.b >> i) & 1
            # create a full adder for the current bit
            full_adder = FullAdder(bit_a, bit_b, carry)
            sum_result |= (full_adder.sum << i)  # set the i-th bit of the sum
            carry = full_adder.carry_out  # update carry for the next bit
        return sum_result, carry


# example usage of the Half Adder, Full Adder, and Eight-Bit Full Adder
if __name__ == "__main__":
    # test half adder
    ha1 = HalfAdder(0, 0)
    print(f"Half Adder (0, 0) => Sum: {ha1.sum}, Carry: {ha1.carry}")  # output: Sum: 0, Carry: 0

    ha2 = HalfAdder(0, 1)
    print(f"Half Adder (0, 1) => Sum: {ha2.sum}, Carry: {ha2.carry}")  # output: Sum: 1, Carry: 0

    ha3 = HalfAdder(1, 0)
    print(f"Half Adder (1, 0) => Sum: {ha3.sum}, Carry: {ha3.carry}")  # output: Sum: 1, Carry: 0

    ha4 = HalfAdder(1, 1)
    print(f"Half Adder (1, 1) => Sum: {ha4.sum}, Carry: {ha4.carry}")  # output: Sum: 0, Carry: 1

    # test full adder
    fa1 = FullAdder(1, 1, 0)
    print(f"Full Adder (1, 1, 0) => Sum: {fa1.sum}, Carry Out: {fa1.carry_out}")  # output: Sum: 0, Carry Out: 1

    fa2 = FullAdder(1, 0, 1)
    print(f"Full Adder (1, 0, 1) => Sum: {fa2.sum}, Carry Out: {fa2.carry_out}")  # output: Sum: 0, Carry Out: 1

    # test eight bit full adder
    eight_bit_adder = EightBitFullAdder(0b11001100, 0b10101010)
    result, carry = eight_bit_adder.add()
    print(f"8-bit Full Adder Result: {bin(result)}, Carry Out: {carry}")  # output: Result in binary, Carry Out: 0 or 1



from datetime import datetime
from typing import Any, Dict, List

import backoff
import pandas as pd
from requests import exceptions, request

from stpstone.utils.cals.handling_dates import DatesBR


class AlphaTools:

    def __init__(
        self,
        str_user: str,
        str_passw: str,
        str_host: str,
        str_instance: str,
        dt_start: datetime,
        dt_end: datetime,
        str_fmt_date_output: str = "YYYY-MM-DD",
    ):
        """
        Connection to INOA Alpha Tools API
        Args:
            str_user (str): username
            str_passw (str): password
            str_host (str): host
            str_instance (str): instance
            dt_start (datetime): start date
            dt_end (datetime): end date
            str_fmt_date_output (str): format date output
        Returns:
            None
        """
        self.str_user = str_user
        self.str_passw = str_passw
        self.str_host = str_host
        self.str_instance = str_instance
        self.dt_start = dt_start
        self.dt_end = dt_end
        self.str_fmt_date_output = str_fmt_date_output

    @backoff.on_exception(
        backoff.constant,
        exceptions.RequestException,
        interval=10,
        max_tries=20,
    )
    def generic_req(
        self, str_method: str, str_app: str, dict_params: dict
    ) -> List[Dict[str, Any]]:
        resp_req = request(
            str_method,
            url=self.str_host + str_app,
            json=dict_params,
            auth=(self.str_user, self.str_passw),
        )
        resp_req.raise_for_status()
        return resp_req.json()

    @property
    def funds(self) -> pd.DataFrame:
        dict_params = {
            "values": [
                "id",
                "name",
                "legal_id",
            ],
            "is_active": None,
        }
        json_req = self.generic_req(
            "POST",
            "funds/get_funds",
            dict_params,
        )
        df_funds = pd.DataFrame.from_dict(json_req, orient="index")
        df_funds = df_funds.astype(
            {
                "id": int,
                "name": str,
                "legal_id": str,
            }
        )
        df_funds["origin"] = self.str_instance
        df_funds.columns = [x.upper() for x in df_funds.columns]
        return df_funds

    def quotes(self, list_ids: List[int]) -> pd.DataFrame:
        dict_params = {
            "fund_ids": list_ids,
            "start_date": self.dt_start.strftime("%Y-%m-%d"),
            "end_date": self.dt_end.strftime("%Y-%m-%d"),
        }
        json_req = self.generic_req(
            "POST",
            "portfolio/get_portfolio_overview_date_range",
            dict_params,
        )
        df_quotes = pd.DataFrame(json_req["items"])
        df_quotes = df_quotes.astype(
            {
                "fund_id": int,
                "date": str,
                "status_display": str,
            }
        )
        df_quotes["date"] = [
            DatesBR().str_date_to_datetime(d, self.str_fmt_date_output)
            for d in df_quotes["date"]
        ]
        df_quotes.columns = [x.upper() for x in df_quotes.columns]
        return df_quotes



### HANDLING NUMERICAL ISSUES ###

import re
import math
import operator
import functools
from fractions import Fraction
from math import gcd
from numbers import Number
from typing import Any, Union, Dict, List, Optional
from stpstone.utils.parsers.str import StrHandler
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class NumHandler(metaclass=TypeChecker):

    def multiples(self, m: int, closest_ceiling_num: int) -> List[Number]:
        """Generate a list of numerical multiples of a given number m, until it reaches or passes
        a given number closest_ceiling_num. The last element of the list will be the closest
        multiple of m to closest_ceiling_num.

        For example, if m = 3 and closest_ceiling_num = 10, the output will be [0, 3, 6, 9].

        Parameters
        ----------
        m : int
            The number of which to generate multiples
        closest_ceiling_num : int
            The number after which to stop generating multiples

        Returns
        -------
        list
            A list of multiples of m, stopping at closest_ceiling_num
        """
        # appending multiples
        list_numerical_mulptiples = list()
        count = int(closest_ceiling_num / m) + 2
        for i in range(0, count * m, m):
            list_numerical_mulptiples.append(i)
        # replacing last value
        if list_numerical_mulptiples[-1] > closest_ceiling_num:
            list_numerical_mulptiples[-1] = closest_ceiling_num
        # output
        return list_numerical_mulptiples

    def nearest_multiple(self, number: Number, multiple: Number) -> int:
        """Return the nearest multiple of a given number. For example, if number = 10.7 and multiple = 3,
        the output will be 9, which is the nearest multiple of 3 to 10.7.
        
        Parameters
        ----------
        number : int or float
            The number for which to find the nearest multiple
        multiple : int
            The number of which to find the nearest multiple
        
        Returns
        -------
        int
            The nearest multiple of multiple to number
        """
        return multiple * int(number / multiple)

    def round_up(self, float_number_to_round: float, float_base: float, float_ceiling: float) \
        -> float:
        """Round a given number up to the nearest multiple of a given base number.
        The new number is capped at a given ceiling.
        
        Parameters
        ----------
        float_number_to_round : float
            The number to round up
        float_base : float
            The base number to round up to
        float_ceiling : float
            The maximum value the new number can take
        
        Returns
        -------
        float
            The rounded up number, capped at the given ceiling
        """
        # correcting variables to float type
        float_number_to_round, float_base, float_ceiling = (float(x) for x in
                                                            [float_number_to_round, float_base, 
                                                             float_ceiling])
        # defining next multiple with a ceiling
        if float(float_base + self.truncate(float_number_to_round / float_base, 0)
                 * float_base) < float_ceiling:
            return float(float_base + self.truncate(float_number_to_round / float_base, 0)
                         * float_base)
        else:
            return float_ceiling

    def decimal_to_fraction(self, decimal_number: Number) -> Fraction:
        """Converts a decimal number to a fraction.

        Parameters
        ----------
        decimal_number : float or str
            The decimal number to convert. Can be a float or a string that 
            represents a decimal number.

        Returns
        -------
        Fraction
            A Fraction object representing the decimal number as a fraction.
        """
        return Fraction(decimal_number)

    def greatest_common_divisor(self, int1: int, int2: int) -> int:
        """Calculate the greatest common divisor (GCD) of two integers.

        Parameters
        ----------
        int1 : int
            The first integer.
        int2 : int
            The second integer.

        Returns
        -------
        int
            The greatest common divisor of int1 and int2.
        """
        return gcd(int1, int2)

    def truncate(self, number: Union[float, int], digits: int) -> float:
        """Truncates a given number to a given number of decimal places.

        Parameters
        ----------
        number : float or int
            The number to truncate.
        digits : int
            The number of decimal places to truncate to.

        Returns
        -------
        float
            The number truncated to the given number of decimal places.
        """
        stepper = 10.0 ** digits
        return math.trunc(stepper * number) / stepper

    def sumproduct(self, *lists: List[List[int]]) -> int:
        """Compute the sum of the products of corresponding elements of a list of lists.
        
        Parameters
        ----------
        *lists : list of lists
            The lists of which to compute the sum of products.
        
        Returns
        -------
        int
            The sum of the products of corresponding elements of the lists.
        """
        return sum(functools.reduce(operator.mul, data) for data in zip(*lists))

    def number_sign(self, number: Union[float, int], base_number: Union[float, int] = 1) -> float:
        """Determine the sign of a number.

        Parameters
        ----------
        number : float or int
            The number whose sign is to be determined.
        base_number : float or int, optional
            The base number whose sign will be returned. Default is 1.

        Returns
        -------
        float
            A float with the magnitude of base_number and the sign of number.
        """
        return math.copysign(base_number, number)

    def multiply_n_elements(self, *args: int) -> int:
        """Multiply the given elements together.

        Parameters
        ----------
        *args : int
            A variable number of integer arguments to be multiplied.

        Returns
        -------
        int
            The product of the given integer arguments.
        """
        product = 1
        for a in args:
            product *= a
        return product

    def sum_n_elements(self, *args: Union[int, float]) -> Union[int, float]:
        """Calculate the sum of a variable number of arguments.

        Parameters
        ----------
        *args : int or float
            A variable number of numerical arguments to be summed.

        Returns
        -------
        int or float
            The sum of the provided arguments.
        """
        sum_ = 0
        for a in args:
            sum_ += a
        return sum_

    def factorial(self, n: int) -> int:
        """Calculate the factorial of a given positive integer n.

        Parameters
        ----------
        n : int
            A positive integer for which the factorial is to be calculated.

        Returns
        -------
        int
            The factorial of the input integer n.
        """
        return functools.reduce(operator.mul, range(1, n + 1))

    def range_floats(self, float_epsilon, float_inf, float_sup, float_pace):
        """Generate a list of float values from float_inf to float_sup, with a step size of
        float_pace. The list is generated by multiplying each of the input parameters by
        float_epsilon, and then dividing each element of the list by float_epsilon. This
        allows the method to be used with any unit of measurement, such as days, hours,
        minutes, etc.

        Parameters
        ----------
        float_epsilon : float
            The unit of measurement for the output list.
        float_inf : float
            The lower bound of the range.
        float_sup : float
            The upper bound of the range.
        float_pace : float
            The step size of the range.

        Returns
        -------
        list
            A list of float values from float_inf to float_sup, with a step size of
            float_pace.
        """
        return [float(x) / float_epsilon for x in range(int(float_inf * float_epsilon),
                                                        int(float_sup * float_epsilon),
                                                        int(float_pace * float_epsilon))]

    def clamp(self, n: float, minn: float, maxn: float) -> float:
        """Clamp a number n to the range [minn, maxn].
        
        Parameters
        ----------
        n : float
            The number to clamp.
        minn : float
            The lower bound of the range.
        maxn : float
            The upper bound of the range.
        
        Returns
        -------
        float
            The clamped number.
        """
        return max(min(maxn, n), minn)

    def is_numeric(self, str_: str) -> bool:
        """Check whether a given string is a valid number.

        Parameters
        ----------
        str_ : str
            The string to check.

        Returns
        -------
        bool
            True if the string is a valid number, False otherwise.
        """
        try:
            float(str_)
            return True
        except ValueError:
            return False

    def is_number(self, value_: Any) -> bool:
        """Check whether a given value is a number (int, float, etc.).

        Parameters
        ----------
        value_ : Any
            The value to check.

        Returns
        -------
        bool
            True if the value is a number, False otherwise.
        """
        return isinstance(value_, (Number)) and not isinstance(value_, bool)

    def transform_to_float(
        self,
        value_: Union[str, int, float, bool], 
        int_precision: Optional[int] = None
    ) -> Union[float, str, bool]:
        """Convert a given value to a float, handling a variety of possible formats and
        edge cases.

        Parameters
        ----------
        value_ : Union[str, int, float, bool]
            The value to convert.
        int_precision : Optional[int]
            The number of decimal places to round to. If None, the original precision is
            preserved.

        Returns
        -------
        Union[float, str, bool]
            The converted float value, or the original value if conversion fails.

        Notes
        -----
        This method is designed to handle a wide range of possible number formats, including
        European and American number formats, percentages, basis points, and others. It also
        handles negative numbers and edge cases like empty strings and non-numeric input.

        Examples
        --------
        >>> transform_to_float('3,132.45%')
        31.3245
        >>> transform_to_float('1.234,56')
        1234.56
        >>> transform_to_float('1,234,567.89')
        1234567.89
        >>> transform_to_float('4.56 bp')
        0.000456
        >>> transform_to_float('(-)912.412.911,231')
        -912412911.231
        >>> transform_to_float(True)
        True
        >>> transform_to_float('Text Tried')
        'Text Tried'
        >>> transform_to_float('-0,10%')
        -0.001
        >>> transform_to_float('0,10%')
        0.001
        >>> transform_to_float('9888 Example')
        '9888 Example'
        """
        # return boolean values as-is
        if isinstance(value_, bool):
            return value_
        # handle non-string cases
        if isinstance(value_, (int, float)):
            return round(value_, int_precision) if int_precision is not None else float(value_)
        original = str(value_).strip()
        s = original
        # first check if this is clearly a mixed number/text case that should remain unchanged
        if re.search(r'[a-zA-Z].*\d|\d.*[a-zA-Z]', s) and not (
            '%' in s or re.search(r'(bp|b\.p\.?)$', s, flags=re.IGNORECASE)):
            return original
        # check for percentage or basis points
        bl_percentage = '%' in s
        bl_bp = re.search(r'(bp|b\.p\.?)$', s, flags=re.IGNORECASE) is not None
        # handle negative numbers (check original string)
        bl_negative = re.search(r'\(.*\)|^-', original.strip()) is not None
        # remove percentage signs, parentheses, and + signs but keep negative signs
        s = re.sub(r'[%\+\(\)]', '', s).strip()
        s = s.replace(')', '')  # remove closing parentheses if any
        # remove only known suffixes (bp, b.p.) but preserve other text
        if bl_bp:
            s_clean = re.sub(r'(bp|b\.p\.?)$', '', s, flags=re.IGNORECASE).strip()
        else:
            s_clean = s
        # count separators to determine format
        comma_count = s_clean.count(',')
        dot_count = s_clean.count('.')
        # handle number formatting
        if comma_count == 1 and dot_count > 0:
            # if there's exactly one comma and it comes after a dot, assume European format
            if s_clean.find(',') > s_clean.find('.'):
                s_clean = s_clean.replace('.', '').replace(',', '.')
            else:
                # otherwise assume American format (1,234.56)
                s_clean = s_clean.replace(',', '')
        elif comma_count > 1:
            # multiple commas - assume American thousands separator (1,234,567.89)
            s_clean = s_clean.replace(',', '')
        elif dot_count > 1:
            # multiple dots - could be European thousands separator (1.234.567,89)
            parts = s_clean.split('.')
            if all(len(p) == 3 for p in parts[:-1]):
                # all parts except last have length 3 → thousands separators
                s_clean = s_clean.replace('.', '')
            else:
                # might be decimal points, but unclear - try removing all dots
                s_clean = s_clean.replace('.', '')
        elif comma_count == 1 and dot_count == 0:
            # single comma - could be European decimal
            # replace comma with decimal point
            s_clean = s_clean.replace(',', '.')
        # try to convert to float
        try:
            num = float(s_clean)
            if bl_negative:
                num = -abs(num)
            # apply percentage or basis point conversion
            if bl_percentage:
                num /= 100
            elif bl_bp:
                num /= 10_000
            if int_precision is not None:
                num = round(num, int_precision)
            return num
        except ValueError:
            return original


import pandas as pd
from datetime import datetime
from typing import Optional
from sqlalchemy.orm import Session
from logging import Logger
from requests import Response
from stpstone._config.global_slots import YAML_DEBENTURES
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.connections.netops.proxies.managers.free import YieldFreeProxy
from stpstone.ingestion.abc.requests import ABCRequests


class DebenturesComBR(ABCRequests):
    """
    Debentures MTM ingestion
    Metadata:
        - https://www.debentures.com.br/exploreosnd/exploreosnd.asp
    Special thanks to Rodrigo Prado (https://github.com/royopa) for helping to develop this class
    """

    def __init__(
        self,
        session: Optional[Session] = None,
        dt_start:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 10),
        dt_end:datetime=DatesBR().sub_working_days(DatesBR().curr_date, 1),
        cls_db:Optional[Session]=None,
        logger:Optional[Logger]=None
    ) -> None:
        self.session = session
        self.dt_start = dt_start
        self.dt_end = dt_end
        self.cls_db = cls_db
        self.logger = logger
        self.dt_ref = dt_end
        self.dt_inf_yyyymmdd = dt_start.strftime('%Y%m%d')
        self.dt_sup_yyyymmdd = dt_end.strftime('%Y%m%d')
        self.dt_inf_ddmmyyyy = dt_start.strftime('%d/%m/%Y')
        self.dt_sup_ddmmyyyy = dt_end.strftime('%d/%m/%Y')
        super().__init__(
            dict_metadata=YAML_DEBENTURES,
            session=session,
            dt_ref=dt_end,
            cls_db=cls_db,
            logger=logger
        )

    def req_trt_injection(self, resp_req:Response) -> Optional[pd.DataFrame]:
        return None



### DIRECTORY AND FILES MANAGEMENT ###

import os
import re
import shutil
import tempfile
import pycurl
import os
import fnmatch
import wget
import py7zr
import hashlib
import tarfile
import chardet
from datetime import datetime
from zipfile import ZipFile, ZIP_DEFLATED
from io import BytesIO, TextIOWrapper, BufferedReader
from typing import Tuple, List, Union, Iterable, Optional
from requests import Response
from pathlib import Path
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.parsers.dicts import HandlingDicts


class DirFilesManagement:

    @property
    def get_cur_dir(self):
        """
        DOCSTRING: GET CURRENT DIRECTORY
        INPUTS: -
        OUTPUTS: CURRENT DIRECTORY
        """
        return os.getcwd()

    def list_dir_files(self, dir_path=None):
        """
        DOCSTRING: RETURN SUBFOLDERS OR FILE NAMES
        INPUTS: DIR NAME (IN CASE THIS ARGUMENT IS NONE THE RETURNED VALUE
        IS FILES IN THE FOLDER)
        OUTPUTS: LIST OF FILES OR SUBFOLDERS
        """
        return os.listdir(dir_path)

    def change_dir(self, dir_path):
        """
        DOCSTRING: CHANGE CURRENT DIRECTORY
        INPUTS: DIRECTORY NAME
        OUTPUTS: -
        """
        os.chdir(dir_path)

    def mk_new_directory(self, dir_path):
        """
        DOCSTRING: MAKE A NEW DIRECTORY
        INPUSTS: NAME OF THE DIRECTORY
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)
            return True
        else:
            return False

    def move_file(self, old_file_name, new_file_name):
        """
        DOCSTRING: MOVE A FILE FROM ORIGINAL DIRECTORY TO ANOTHER (IT DELETES THE OLD ONE)
        INPUTS: OLD AND NEW COMPLETE PATH NAME, AND DELETE OLD ONE
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        shutil.move(old_file_name, new_file_name)
        return self.object_exists(new_file_name)

    def rename_dir_file(self, old_object_name, new_object_name):
        """
        DOCSTRING: RENAMING FILES OR FOLDERS
        INPUTS: OLD AND NEW COMPLETE PATH
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        # renaming
        os.rename(old_object_name, new_object_name)
        # return status of accomplishment
        if os.path.exists(new_object_name):
            return True
        else:
            return False

    def removing_dir(self, dir_path):
        """
        DOCSTRING: REMOVE A DIRECTORY
        INPUTS: COMPLETE PATH OF THE DIRECTORY
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        if len(self.list_dir_files(dir_path)) == 0:
            os.rmdir(dir_path)
        else:
            shutil.rmtree(dir_path)
        if not os.path.exists(dir_path):
            return True
        else:
            return False

    def removing_file(self, file_path):
        """
        DOCSTRING: REMOVE A FILE
        INPUTS: COMPLETE NAME
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        os.remove(file_path)
        if not os.path.exists(file_path):
            return True
        else:
            return False

    def object_exists(self, object_path):
        """
        DOCSTRING: BLAMES WHETER OR NOT FILE/FOLDER HAS BEEN CREATED
        INPUTS: OBJECT PATH
        OUTPUTS: OK/NOK
        """
        if os.path.exists(object_path):
            return True
        else:
            return False

    def time_last_edition(self, object_path, bl_to_datetime=False):
        """
        DOCSTRING: TIMESTAMP WITH LAST SAVED EDITION IN THE FILE
        INPUTS: OBJECT PATH
        OUTPUTS: TUPLE WITH TIMESTAMP OF LAST EDITION AND WHETER FILE EXISTS OR NOT
        """
        if os.path.exists(object_path):
            if bl_to_datetime == True:
                return (datetime.fromtimestamp(os.path.getmtime(object_path)), True)
            else:
                return (os.path.getmtime(object_path), True)
        else:
            return ('INTERNAL ERROR', False)

    def time_creation(self, object_path):
        """
        DOCSTRING: TIMESTAMP WITH CREATION OF FILE
        INPUTS: OBJECT PATH
        OUTPUTS: TUPLE WITH TIMESTAMP OF FILE CREATION AND WHETER FILE EXISTS OR NOT
        """
        if os.path.exists(object_path):
            return (os.path.getctime(object_path), True)
        else:
            return ('INTERNAL ERROR', False)

    def time_last_access(self, object_path):
        """
        DOCSTRING: TIMESTAMP WITH LAST ACCESS TO THE FILE
        INPUTS: OBJECT PATH
        OUTPUTS: TUPLE WITH TIMESTAMP OF FILE LAST ACCESS AND WHETER FILE EXISTS OR NOT
        """
        if os.path.exists(object_path):
            return (os.path.getatime(object_path), True)
        else:
            return ('INTERNAL ERROR', False)

    def get_file_name_path_split(self, complete_file_name):
        """
        DOCSTRING: GET FILE PATH AND NAME IN A TUPLE
        INPUTS: COMPLETE FILE NAME
        OUTPUT: RETURNS TUPLE WITH FILE NAME HEAD (PATH) AND TAIL (NAME)
        """
        return os.path.split(complete_file_name)

    def join_n_path_components(self, *path_components):
        """
        DOCSTRING: JOIN PATH COMPONENTS
        INPUTS: N-PATH COMPONENTS
        OUTPUTS: OUTPUT COMPLETE PATH
        """
        path_output = ''
        for path_component in path_components:
            path_output = os.path.join(path_output, path_component)
        return path_output

    def get_filename_parts_from_url(self, url):
        """
        DOCSTRING: GET FILE NAME FROM A COMPLETE URL
        INPUTS: COMPLETE URL
        OUTPUTS: FILENAME WITH AND WITHOUT EXTENSION, IN STR AND LIST TYPES, RESPECTIVELLY
        """
        fullname = url.split('/')[-1].split('#')[0].split('?')[0]
        t = list(os.path.splitext(fullname))
        if t[1]:
            t[1] = t[1][1:]
        return t

    def get_file_extensions(self, file_path: str) -> List[Union[str, None]]:
        return re.findall(r'\.([a-zA-Z0-9_]+)(?:[\?#]|$)', file_path)

    def get_last_file_extension(self, file_path: str) -> str:
        list_ = self.get_file_extensions(file_path)
        if len(list_) > 0:
            return list_[-1]
        else:
            return None

    def download_web_file(self, url, filepath=None):
        """
        DOCSTRING: DOWNLOAD FILE FROM WEB (DOWNLOADED TEMPORARY FILENAME IF NO FILEPATH IS PROVIDED)
        INPUTS: COMPLETE PATH TO FILE AND FILE NAME WITH EXTENSION
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        OBS: IF IT IS NEEDED TO PASS CREDENTIALS PARAMETER URL_PATH OUGHT BE DECLARED AS:
            'ftp://username:password@server/path/to/file'
        """
        # removing previous version
        if self.object_exists(filepath) == True:
            _ = self.removing_file(filepath)
        # fetching data
        if not filepath:
            _, suffix = self.get_filename_parts_from_url(url)
            f = tempfile.NamedTemporaryFile(suffix='.' + suffix, delete=False)
            filepath = f.name
        else:
            f = open(filepath, 'wb')
        c = pycurl.Curl()
        c.setopt(pycurl.URL, str(url))
        c.setopt(pycurl.WRITEFUNCTION, f.write)
        try:
            c.perform()
            c.close()
            f.close()
            return self.object_exists(filepath)
        except:
            c.close()
            f.close()
            _ = self.removing_file(filepath)
            wget.download(url, filepath)
            return self.object_exists(filepath)

    def unzip_files_from_dir(self, destination_path):
        """
        DOCSTRING: UNZIP ALL FILES FROM A FOLDER
        INPUTS: DESTINATION PATH
        OUTPUTS: NONE
        """
        list_files_unz = list()
        files = os.listdir(destination_path)
        for file in files:
            if file.endswith('.zip'):
                filePath = destination_path + '/' + file
                zip_file = ZipFile(filePath)
                list_files_unz.append(zip_file.namelist())
                for names in zip_file.namelist():
                    zip_file.extract(names, destination_path)
                zip_file.close()
        return list_files_unz

    def unzip_file(self, zippedfile_path, dir_destiny):
        """
        DOCSTRING: UNZIP ONE SINGULAR ZIP FILE TO A DESTINATION
        INPUTS: ZIPPED FILE PATH AN DESTINATION PATH
        OUTPUTS: LIST OF UNZIPPED FILES
        """
        with ZipFile(zippedfile_path, 'r') as zipobj:
            list_zip_files = zipobj.namelist()
            zipobj.extractall(dir_destiny)
        return list_zip_files

    def compress_to_zip(self, list_files_archive, zfilename):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        # creating object of zipfile compression
        zout = ZipFile(zfilename, 'w', ZIP_DEFLATED)
        # looping through archive files
        for fname in list_files_archive:
            zout.write(fname)
        return True

    def compress_to_7z_file(self, file_path_7z, object_to_compress, method='w'):
        """
        REFERENCES: https://github.com/miurahr/py7zr
        DOCSTRING: ZIP FILE TO 7ZR, OBJECT TO COMPRESS PATH, AND METHOD (WRTIE AS DEFAULT)
        INPUTS: 7ZR FILE NAME, DESTINY DIRECTORY
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        """
        with py7zr.SevenZipFile(file_path_7z, mode=method) as archive:
            archive.writeall(object_to_compress)
        return self.object_exists(file_path_7z)

    def decompress_7z_file(self, file_path_7z, method='r'):
        """
        REFERENCES: https://github.com/miurahr/py7zr
        DOCSTRING: ZIP FILE TO 7ZR
        INPUTS: 7ZR FILE NAME, STR ARCNAME (BASE AS DEFAULT) AND STR MODE (READ AS DEFAULT)
        OUTPUTS: LIST OF FILE NAMES COMPRESSED
        """
        with py7zr.SevenZipFile(file_path_7z, mode=method) as archive:
            list_file_names = archive.getnames()
            archive.extractall()
        return list_file_names

    def choose_last_saved_file_w_rule(self, parent_dir, name_like):
        """
        DOCSTRING: CHOOSE LAST SAVED FILE WITH RULE
        INPUTS: PARENT DIR AND PART OF THE NAME OF THE FILE
        OUTPUTS: NOK OR COMPLETE NAME OF THE FILE
        """
        # setting passaging variables
        files_dir = os.listdir(parent_dir)
        file_dir = None
        file_name_return = None
        # looping through all files in the folder and returning the last edited one with the
        #   name like given
        for file_dir in files_dir:
            if fnmatch.fnmatch(file_dir, name_like):
                if file_name_return is None:
                    file_name_return = file_dir
                else:
                    if os.path.getmtime(parent_dir
                                        + file_dir) > os.path.getmtime(
                                            os.path.join(parent_dir, file_name_return)):
                        file_name_return = file_dir
        # return the complete file path, or NOK whether it has not been found
        if file_name_return is None:
            return False
        else:
            return os.path.join(parent_dir, file_name_return)

    def copy_file(self, org_file_path, dest_direcory):
        """
        DOCSTRING: COPY FILE TO A FOLDER
        INPUTS: ORIGINAL AND DESTINATION COMPLETE FILE PATH
        OUTPUTS: STATUS OF ACCOMPLISHMENT - NO ORIGINAL FILE/OK
        """
        if os.path.exists(org_file_path):
            shutil.copy(org_file_path, dest_direcory)
            return True
        else:
            return 'NO ORIGINAL FILE'

    def walk_folder_subfolder_w_rule(self, root_directory, list_name_like):
        """
        DOCSTRING: WALK THROUGH ALL FILES IN A FOLDER AND ITS SUBFOLDERS, RETURNING COMPLETE PATH OF
            FILES WITH A NAME LIKE OF INTEREST
        INPUTS: ROOT DIRECTORY
        OUTPUTS: LIST OF FILE PATHS
        """
        list_paths = list()
        for directory, _, files in os.walk(root_directory):
            for file in files:
                if any([fnmatch.fnmatch(file, name_like) == True for name_like in list_name_like]):
                    list_paths.append(os.path.join(directory, file))
        return list_paths

    def walk_folder_subfolder(self, root_directory):
        """
        DOCSTRING: WALK THROUGH ALL FILES IN A FOLDER AND ITS SUBFOLDERS
        INPUTS: ROOT DIRECTORY
        OUTPUTS: LIST OF FILES PATHS
        """
        list_paths = list()
        for directory, _, files in os.walk(root_directory):
            for file in files:
                list_paths.append(os.path.join(directory, file))
        return list_paths

    def loop_files_w_rule(self, directory, name_like, bl_first_last_edited=True,
                          bl_to_datetime=True, key_file_name='file_name',
                          key_file_last_edition='file_last_edition'):
        """
        DOCSTRING: RETURN FILES FROM A FOLDER WITH A GIVEN RULE
        INPUTS: DIRECTORY AND RULE (NAME_LIKE)
        OUTPUTS: RETURNS FILES PATHS WITH A GIVEN RULE
        """
        # creating list of files in a given directory with a given part of name
        list_files_names_like = [file_name for file_name in os.listdir(directory)
                                 if StrHandler().match_string_like(file_name, name_like)]
        # checking whether it is necessary to retrieve the files in a last edition order
        if bl_first_last_edited == False:
            return list_files_names_like
        else:
            #   list of last edition times
            list_files_last_edition = [self.time_last_edition(os.path.join(
                directory, file_name), bl_to_datetime=bl_to_datetime)
                for file_name in list_files_names_like]
            #   creating a list of dictionaries with name and last edition time
            list_ser_file_name_last_edition = [{
                key_file_name: list_files_names_like[i],
                key_file_last_edition: list_files_last_edition[i][0]
            } for i in range(len(list_files_last_edition))]
            #   sort list of dictionaries
            return [os.path.join(directory, dict_[key_file_name]) for dict_
                    in HandlingDicts().multikeysort(list_ser_file_name_last_edition,
                                                    ['-' + key_file_last_edition])]

    def list_dir_files(self, dir_path=None):
        """
        DOCSTRING: RETURN SUBFOLDERS OR FILE NAMES
        INPUTS: DIR NAME (IN CASE THIS ARGUMENT IS NONE THE RETURNED VALUE
        IS FILES IN THE FOLDER)
        OUTPUTS: LIST OF FILES OR SUBFOLDERS
        """
        return os.listdir(dir_path)

    def find_project_root(self, marker:str='pyproject.toml') -> Path:
        """
        Traverse up the directory tree to find the project root
        by looking for a marker file (e.g., pyproject.toml, README.md, .git).
        """
        current_path = Path(__file__).resolve()
        while current_path != current_path.parent:  # Stop at the filesystem root
            if (current_path / marker).exists():
                return current_path
            current_path = current_path.parent
        raise FileNotFoundError(f"Could not find project root with marker: {marker}")

    def get_file_format_from_file_name(self, filename):
        """
        DOCSTRING: GET FILE FORMAT FROM FILEN NAME
        INPUTS: FILE NAME
        OUTPUTS: FORMAT
        """
        return filename.split('.')[-1]

    def get_file_size(self, filename):
        """
        DOCSTRING: GET FILE SIZE IN BYTES
        INPUTS: FILENAME
        OUTPUTS: FLOAT
        """
        return os.path.getsize(filename)

    def recursive_extract_zip(self, zip_file_path: str, extract_dir: str) -> None:
        """
        Recursively extracts the contents of a ZIP file and any nested ZIP files to a target
        directory, until no ZIP file is available

        Args:
            zip_file_path (str): The path to the ZIP file to be extracted.
            extract_dir (str): The directory where the contents of the ZIP file should be extracted.

        Returns:
            None
        """
        with ZipFile(zip_file_path, "r") as zip_ref:
            zip_ref.extractall(extract_dir)
        os.remove(zip_file_path)
        for root_path, _, files in os.walk(extract_dir):
            for file in files:
                if file.endswith(".zip"):
                    nested_zip_path = os.path.join(root_path, file)
                    self.recursive_extract_zip(nested_zip_path, root_path)


class RemoteFiles(DirFilesManagement):

    def get_file_from_zip(
        self,
        resp_req: Response,
        path_dir: Union[str, tempfile.TemporaryDirectory, Path],
        tup_endswith: Tuple[str]
    ) -> str:
        zip_file_path = os.path.join(path_dir, "archive.zip")
        with open(zip_file_path, "wb") as zip_file:
            zip_file.write(resp_req.content)
        self.recursive_extract_zip(zip_file_path, path_dir)
        ex_file_path = None
        for root_path, _, list_files in os.walk(path_dir):
            for file in list_files:
                if file.endswith(tup_endswith):
                    ex_file_path = os.path.join(root_path, file)
                    break
            if ex_file_path:
                break
        if not ex_file_path:
            raise ValueError("No file found in the extracted .zip archive. "
                             + f"- considerered extensions: {tup_endswith}")
        return ex_file_path

    def get_zip_from_web_in_memory(
        self,
        resp_req: Response,
        bl_io_interpreting: bool = False
    ) -> Union[TextIOWrapper, BufferedReader, List[BufferedReader]]:
        zipfile = ZipFile(BytesIO(resp_req.content))
        zip_names = zipfile.namelist()
        if len(zip_names) == 1:
            file_name = zip_names.pop()
            extracted_file = zipfile.open(file_name)
            if bl_io_interpreting == True:
                return TextIOWrapper(extracted_file)
            else:
                return extracted_file
        return [zipfile.open(file_name) for file_name in zip_names]

    def calculate_file_hash(file_path: str, algorithm: str = "sha256") -> str:
        hash_func = getattr(hashlib, algorithm)()
        with open(file_path, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                hash_func.update(chunk)
        return hash_func.hexdigest()

    def validate_file_hash(
        self, file_path: Union[str, Path], expected_hash: str, algorithm: str = "sha256") -> bool:
        """
        Validates the integrity of a file by comparing its hash with an expected value.

        Args:
            file_path (Union[str, Path]): The path to the file.
            expected_hash (str): The expected hash value.
            algorithm (str): The hashing algorithm to use (e.g., "md5", "sha256").

        Returns:
            bool: True if the hash matches, False otherwise.
        """
        hash_func = getattr(hashlib, algorithm)()
        with open(file_path, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                hash_func.update(chunk)
        return hash_func.hexdigest() == expected_hash

    def extract_file(self, archive_path: Union[str, Path], extract_dir: Union[str, Path],
                     format: str = "zip") -> bool:
        """
        Extracts files from an archive to a specified directory.

        Args:
            archive_path (Union[str, Path]): The path to the archive file.
            extract_dir (Union[str, Path]): The directory where files should be extracted.
            format (str): The format of the archive (e.g., "zip", "tar", "7z").

        Returns:
            bool: True if extraction was successful, False otherwise.
        """
        try:
            if format == "zip":
                with ZipFile(archive_path, "r") as zip_ref:
                    zip_ref.extractall(extract_dir)
            elif format == "tar":
                with tarfile.open(archive_path, "r:*") as tar_ref:
                    tar_ref.extractall(extract_dir)
            elif format == "7z":
                with py7zr.SevenZipFile(archive_path, mode="r") as seven_zip_ref:
                    seven_zip_ref.extractall(extract_dir)
            else:
                raise ValueError(f"Unsupported archive format: {format}")
            return True
        except Exception as e:
            print(f"Failed to extract archive: {e}")
            return False

    def get_file_metadata(self, file_path: Union[str, Path]) -> dict:
        stat = os.stat(file_path)
        return {
            "size": stat.st_size,
            "creation_time": datetime.fromtimestamp(stat.st_birthtime),
            "modification_time": datetime.fromtimestamp(stat.st_mtime),
            "access_time": datetime.fromtimestamp(stat.st_atime),
        }

    def stream_file(self, resp_req: Response, chunk_size: int = 8192) -> Iterable[bytes]:
        """
        Streams a file from a remote URL in chunks.

        Args:
            url (str): The URL of the file to stream.
            chunk_size (int): The size of each chunk in bytes.

        Returns:
            Iterable[bytes]: An iterable yielding file chunks.
        """
        for chunk in resp_req.iter_content(chunk_size=chunk_size):
            yield chunk

    def check_separator_consistency(
        self,
        req_content: bytes,
        int_skip_rows: int = 0,
        int_skip_footer: int = 0,
        list_sep: Optional[List[str]] = [",", ";", "\t"]
    ) -> bool:
        result = chardet.detect(req_content)
        encoding = result["encoding"] if result["encoding"] is not None else "latin-1"
        decoded_content = req_content.decode(encoding)
        list_lines = decoded_content.splitlines()
        int_skip_footer = len(list_lines) - int_skip_footer
        list_lines = list_lines[int_skip_rows:int_skip_footer]
        for sep in list_sep:
            list_sep_counts = [len(line.split(sep)) for line in list_lines]
            if (len(set(list_sep_counts)) == 1) and (all(x > 1 for x in  list_sep_counts)):
                return True
        return False


class FoldersTree:
    def __init__(self, str_path, bl_ignore_dot_folders=False, list_ignored_folders=None,
                 bl_add_linebreak_markdown=False):
        """
        DOCSTRING: INITIALIZE THE CLASS
        INPUTS: PATH, IGNORE DOT FOLDERS (OPTIONAL), LIST OF IGNORED FOLDERS (OPTIONAL)
        OUTPUTS: -
        """
        self.str_path = str_path
        self.bl_ignore_dot_folders = bl_ignore_dot_folders
        self.list_ignored_folders = list_ignored_folders or ['__pycache__']
        self.bl_add_linebreak_markdown = bl_add_linebreak_markdown

    def generate_tree(self, str_curr_path=None, bl_is_last=True, str_prefix='', bl_include_root=True,
                      str_tree_structure=''):
        """
        DOCSTRING: GENERATE A TREE STRUCTURE OF THE DIRECTORY
        INPUTS: CURRENT PATH (OPTIONAL), IS LAST ENTRY (OPTIONAL), PREFIX (OPTIONAL),
            INCLUDE ROOT (OPTIONAL)
        OUTPUTS: A string representation of the directory tree structure.
        """
        # initializing the tree structure
        if str_curr_path is None:
            str_curr_path = self.str_path
        # line break if bl_add_linebreak_markdown is True
        if self.bl_add_linebreak_markdown == True:
            str_linebreak_md = '<br>'
        else:
            str_linebreak_md = ''
        # add the parent folder name as the first line if bl_include_root is True
        if bl_include_root:
            str_tree_structure += f'{os.path.basename(self.str_path)}{str_linebreak_md}\n'
            #   reset str_prefix for the root folder
            str_prefix = ''
        # sort the entries
        list_entries = sorted(os.listdir(str_curr_path))
        # loop through the entries
        for idx, str_entry in enumerate(list_entries):
            str_entry_path = os.path.join(str_curr_path, str_entry)
            # skip ignored folders
            if self.bl_ignore_dot_folders and str_entry.startswith('.'):
                continue
            if str_entry in self.list_ignored_folders:
                continue
            # creating brach prefix
            bl_is_directory = os.path.isdir(str_entry_path)
            bl_is_last_entry = idx == len(list_entries) - 1
            str_branch_prefix = '└── ' if bl_is_last_entry else '├── '
            str_tree_structure += f'{str_prefix}{str_branch_prefix}{str_entry}{str_linebreak_md}\n'
            # if the str_entry is a directory recursively add subdirectories
            if bl_is_directory:
                # Recursively add subdirectories
                str_new_prefix = str_prefix + ('    ' if bl_is_last_entry else '│   ')
                str_tree_structure += self.generate_tree(
                    str_entry_path,
                    bl_is_last=bl_is_last_entry,
                    str_prefix=str_new_prefix,
                    bl_include_root=False
                )
        # return the tree structure
        return str_tree_structure

    @property
    def print_tree(self):
        """
        DOCSTRING: PRINT THE TREE STRUCTURE
        INPUTS:
        OUTPUTS:
        """
        print(self.generate_tree())

    def export_tree(self, filename=None):
        """
        DOCSTRING: EXPORT THE TREE STRUCTURE TO A FILE
        INPUTS: FILENAME (OPTIONAL)
        OUTPUTS: -
        """
        str_tree_structure = self.generate_tree()
        if filename:
            with open(filename, 'w', encoding='utf-8') as file:
                file.write(str_tree_structure)
            print(f'Tree structure has been written to {filename}')
        else:
            return str_tree_structure



### GENERIC PIPELINES ###

from typing import Any, Callable, List

import pandas as pd


def generic_pipeline(data: Any, functions: List[Callable]) -> Any:
    """
    Applies a sequence of functions to a given data object.
    Args:
        data (Any): Initial data input (pandas DataFrame, string, number, list, etc.).
        functions (List[Callable]): A list of functions to apply sequentially.
    Returns:
        Any: The final processed data after applying all functions.
    """
    for func in functions:
        try:
            #   check if the function is designed for DataFrame and if data is a DataFrame
            if isinstance(data, pd.DataFrame) and callable(func):
                data = func(data)
            elif not isinstance(data, pd.DataFrame):
                data = func(data)
        except Exception as e:
            print(f"Error in {func.__name__}: {e}")
            break
    return data



from stpstone.transformations.validation.metaclass_type_checker import TypeChecker


class BinaryComparator(metaclass=TypeChecker):
    def __init__(self, a: int, b: int) -> None:
        """Initialize the Binary Comparator with two binary inputs."""
        self.a = a
        self.b = b

    @property
    def compare(self) -> str:
        """
        Compare two binary inputs and return a comparison result as a string.

        This property compares the binary inputs `a` and `b` and returns a string
        indicating their relationship. The possible return values are:
            - "A is less than B" if `a` is less than `b`
            - "A is greater than B" if `a` is greater than `b`
            - "A is equal to B" if `a` is equal to `b`

        Returns:
            str: A string describing the comparison result between `a` and `b`.
        """
        if self.a < self.b:
            return "A is less than B"
        elif self.a > self.b:
            return "A is greater than B"
        else:
            return "A is equal to B"


if __name__ == "__main__":
    comparator = BinaryComparator(0b1100, 0b1010)
    result = comparator.compare()
    # output: A is greater than B
    print(f"Binary Comparator (1100, 1010) => {result}")



from collections import deque
from typing import Any


class Queue:

    def __init__(self):
        self._items = deque()

    @property
    def is_empty(self) -> bool:
        return not self._items

    def enqueue(self, item: Any) -> None:
        self._items.appendleft(item)

    @property
    def dequeue(self):
        if self.is_empty:
            raise IndexError("Dequeue from an empty queue")
        return self._items.pop()

    @property
    def size(self) -> int:
        return len(self._items)



### HANDLING PICKLE OBJECTS ###

import pickle
import os


class PickleFiles:

    def dump_message(self, message, pickle_file, pickle_protocol=pickle.HIGHEST_PROTOCOL,
                     pickle_protocol_error=-1):
        """
        REFERENCES: https://stackoverflow.com/questions/11218477/how-can-i-use-pickle-to-save-a-dict-or-any-other-python-object
        DOCSTRING: DUMP MESSAGE AND SAVE TO SPECIFIC DIRECTORY IN THE NETWORK
        INPUTS: MESSAGE (STR IN DICT FORMAT), PICKLE FILE AS A COMPLETE FILE .PICKLE NAME
        OUTPUTS: WHETER FILE SAVING WAS SUCCESFUL OR NOT (OK, NOK)
        """
        # store data
        with open(pickle_file, 'wb') as write_file:
            try:
                pickle.dump(message, write_file, protocol=pickle_protocol)
            except:
                pickle.dump(message, write_file,
                            protocol=pickle_protocol_error)
        # return status of accomplishment
        if not os.path.exists(pickle_file):
            return 'NOK'
        else:
            return 'OK'

    def load_message(self, pickle_file, encoding=None, decoding=None):
        """
        REFERENCES: https://stackoverflow.com/questions/11218477/how-can-i-use-pickle-to-save-a-dict-or-any-other-python-object
        DOCSTRING: LOAD MESSAGE FROM PICKLE
        INPUTS: PICKLE FILE
        OUTPUTS: DATA IN MEMORY
        """
        if encoding is None:
            with open(pickle_file, 'rb') as read_file:
                return pickle.load(read_file)
        else:
            with open(pickle_file, encoding=encoding) as read_file:
                return pickle.loads(read_file.read().encode(encoding).decode(decoding))



import pandas as pd
import numpy as np
from requests import Session
from logging import Logger
from numbers import Number
from pydantic import validate_arguments, ConfigDict
from typing import get_type_hints, get_origin, get_args, Type, Dict, Any, Union, BinaryIO, IO
from io import BytesIO, RawIOBase, BufferedIOBase


class TypeChecker(type):
    def __new__(cls: Type["TypeChecker"], name: str, bases: tuple, dict_: Dict[str, Any]) \
        -> "TypeChecker":
        for attr_name, attr_value in dict_.items():
            if (callable(attr_value)) and (not attr_name.startswith("__")):
                bl_arbitrary_types = False
                tup_types_ignore = (pd.DataFrame, np.ndarray, pd.Series, list, Session, Logger, 
                                    Number)
                io_types = (IO, BinaryIO, RawIOBase, BufferedIOBase, BytesIO)
                type_hints = get_type_hints(attr_value)
                for hint in type_hints.values():
                    if hint in tup_types_ignore:
                        bl_arbitrary_types = True
                        break
                    if hint in io_types:
                        bl_arbitrary_types = True
                        break
                    if get_origin(hint) is Union:
                        args = get_args(hint)
                        if any(arg in tup_types_ignore for arg in args):
                            bl_arbitrary_types = True
                            break
                    if getattr(hint, "__origin__", None) in tup_types_ignore:
                        bl_arbitrary_types = True
                        break
                dict_[attr_name] = validate_arguments(
                    attr_value,
                    config=ConfigDict(arbitrary_types_allowed=True) if bl_arbitrary_types else {},
                )
        return super().__new__(cls, name, bases, dict_)



# HANDLING INTERPOLATION LIBRARIES FOR PYTHON

import numpy as np
from scipy.interpolate import interp1d, CubicSpline, lagrange


class Interpolation:

    def set_as_array(self, array_):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return np.array(array_)

    def linear_interpolation(self, array_x, array_y):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return interp1d(self.set_as_array(array_x), self.set_as_array(array_y))

    def cubic_spline(self, array_x, array_y, bc_type='natural'):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return CubicSpline(self.set_as_array(array_x), self.set_as_array(array_y),
                           bc_type=bc_type)

    def lagrange(self, array_x, array_y):
        """
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        """
        return lagrange(self.set_as_array(array_x), self.set_as_array(array_y))

    def divided_diff_newton_polynomial_interpolation(self, array_x, array_y):
        """
        REFERENCES: PYTHON PROGRAMMING AND NUMERICAL METHODS - A GUIDE FOR ENGINEERS AND SCIENTISTS;
            QINGKAI KON, TIMMY SIAUW AND ALEXANDRE M. BAYERN
            https://pythonnumericalmethods.berkeley.edu/notebooks/chapter17.05-Newtons-Polynomial-Interpolation.html
        DOCSTRING: FUNCTION TO CALCULATE THE DIVIDED DIFFERENCE TABLE
        INPUTS: ARRAY INDEPENDENT AND DEPENDENT
        OUTPUTS: ARRAY
        """
        # setting initial parameters
        n = len(array_y)
        matrix_coef = np.zeros([n, n])
        # the first column is array_y
        matrix_coef[:, 0] = array_y
        # looping through i and j to build the coefficients
        for j in range(1, n):
            for i in range(n - j):
                matrix_coef[i][j] = (matrix_coef[i + 1][j - 1] - matrix_coef[i][j - 1]) \
                    / (array_x[i + j] - array_x[i])
        return matrix_coef

    def newton_polynomial_interpolation(self, array_x, array_y,
                                        array_x_range):
        """
        DOCSTRING: EVALUATE THE NEWTON POLYNOMIAL AT NEW DATA (ARRAY INDEP NEW)
        INPUTS: ARRAY INDEPENDENT, DEPENDENT (OF GIVEN DATA POINTS POINTS) AND ARRAY OF X RANGE
        OUTPUTS: ARRAY OF Y RANGE
        """
        # length of independent old array
        n = len(array_x) - 1
        # matrix of divided different coefficients
        matrix_coef = self.divided_diff_newton_polynomial_interpolation(
            array_x, array_y)
        # setting initial parameters of new dependent array
        array_y_range = matrix_coef[n]
        # looping through all data of array independent
        for k in range(1, n + 1):
            #   creating y axis to plotted
            array_y_range = matrix_coef[n - k] + (array_x_range - array_x[
                n - k]) * array_y_range
        return array_y_range



import numpy as np
import pandas as pd
from logging import Logger
from typing import Any, Dict, List, Optional
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from stpstone._config.global_slots import YAML_GEN
from stpstone.transformations.validation.metaclass_type_checker import TypeChecker
from stpstone.utils.cals.handling_dates import DatesBR
from stpstone.utils.loggs.create_logs import CreateLog
from stpstone.utils.parsers.folders import DirFilesManagement
from stpstone.utils.parsers.lists import ListHandler
from stpstone.utils.parsers.str import StrHandler
from stpstone.utils.pipelines.generic import generic_pipeline


pd.set_option('future.no_silent_downcasting', True)

class DFStandardization(metaclass=TypeChecker):

    def __init__(
        self,
        dict_dtypes: Dict[str, Any],
        cols_from_case: Optional[str] = None,
        cols_to_case: Optional[str] = None,
        list_cols_drop_dupl: List[str] = None,
        dict_fillna_strt: Optional[Dict[str, str]] = {},
        str_fmt_dt: str = "YYYY-MM-DD",
        type_error_action: str = "raise",
        strategy_keep_when_dupl: str = "first",
        str_data_fillna: str = "-99999",
        str_dt_fillna: Optional[str] = None,
        str_tz: str = "UTC",
        encoding: str = "latin-1",
        bl_debug: bool = False,
        logger: Optional[Logger] = None,
    ):
        self.dict_dtypes = dict_dtypes
        self.cols_from_case = cols_from_case
        self.cols_to_case = cols_to_case
        self.list_cols_drop_dupl = list_cols_drop_dupl
        self.dict_fillna_strt = dict_fillna_strt
        self.str_fmt_dt = str_fmt_dt
        self.type_error_action = type_error_action
        self.strategy_keep_when_dupl = strategy_keep_when_dupl
        self.str_data_fillna = str_data_fillna
        self.str_dt_fillna = str_dt_fillna if str_dt_fillna is not None else (
            '2099-12-31' if self.str_fmt_dt == 'YYYY-MM-DD' else
            '31/12/2099' if self.str_fmt_dt == 'DD/MM/YYYY' else
            '31/12/99' if self.str_fmt_dt == 'DD/MM/YY' else
            '31122099' if self.str_fmt_dt == 'DDMMYYYY' else
            '311299' if self.str_fmt_dt == 'DDMMYY' else
            '20991231' if self.str_fmt_dt == 'YYYYMMDD' else
            4102358400 if self.str_fmt_dt == 'unix_ts' else
            '31122099'
        )
        self.list_cols_dt = [key for key, value in dict_dtypes.items() if value == "date"]
        self.str_tz = str_tz
        self.encoding = encoding
        self.bl_debug = bl_debug
        self.logger = logger

    def check_if_empty(self, df_: pd.DataFrame) -> pd.DataFrame:
        if df_.empty == True:
            raise ValueError("DataFrame is empty")
        return df_

    def clean_encoding_issues(self, df_: pd.DataFrame) -> pd.DataFrame:
        """
        Cleans encoding issues in a DataFrame by removing non-decodable characters.

        Args:
            df_ (pd.DataFrame): The DataFrame to clean.

        Returns:
            pd.DataFrame: A DataFrame with encoding issues removed.
        """
        def clean_cell(x):
            if isinstance(x, str):
                # remove non-printable and non-ASCII characters
                x = "".join(
                    char for char in x if char.isprintable() and ord(char) < 128
                )
                # replace problematic characters with a placeholder
                x = x.encode(self.encoding, errors="ignore").decode(self.encoding)
            return x
        return df_.apply(clean_cell)

    def column_names(self, df_: pd.DataFrame) -> List[str]:
        if all([x is not None for x in [self.cols_from_case, self.cols_to_case]]):
            list_cols = [
                StrHandler().convert_case(
                    StrHandler().remove_diacritics(col_),
                    self.cols_from_case,
                    self.cols_to_case,
                )
                for col_ in list(df_.columns)
            ]
        else:
            list_cols = list(df_.columns)
        df_.columns = list_cols
        return df_

    def limit_columns_to_dtypes(self, df_: pd.DataFrame) -> pd.DataFrame:
        """
        Limit the DataFrame columns to the keys of self.dict_dtypes

        Args:
            df_ (pd.DataFrame): DataFrame to filter

        Returns:
            pd.DataFrame
        """
        if self.logger is not None:
            CreateLog().info(
                self.logger,
                f"List cols dataframe before filtering: {list(df_.columns)}",
            )
            CreateLog().info(
                self.logger, f"List cols to filter: {list(self.dict_dtypes.keys())}"
            )
            CreateLog().info(
                self.logger,
                "List of columns excluded: "
                + f"{[x for x in list(df_.columns) if x not in list(self.dict_dtypes.keys())]}",
            )
        if self.bl_debug == True:
            print(f"List cols dataframe before filtering: {list(df_.columns)}")
            print(f"List cols to filter: {list(self.dict_dtypes.keys())}")
            print(
                "List of columns excluded: "
                + f"{[x for x in list(df_.columns) if x not in list(self.dict_dtypes.keys())]}"
            )
        list_cols = list(self.dict_dtypes.keys())
        df_ = df_[list_cols]
        return df_

    def delete_empty_rows(self, df_: pd.DataFrame) -> pd.DataFrame:
        list_mask = df_.apply(lambda row: row.isin([None, ""]).all(), axis=1)
        df_cleaned = df_[~list_mask]
        return df_cleaned

    def completeness(self, df_: pd.DataFrame) -> pd.DataFrame:
        """
        Completeness - ensure that all required fields have valid (non-null) entries

        Args:
            df_ (pandas.DataFrame): DataFrame to fill

        Returns:
            pd.DataFrame
        """
        df_.replace("", np.nan, inplace=True)
        list_cols = [
            col_ for col_ in list(df_.columns) if col_ not in self.list_cols_dt
        ]
        if self.dict_fillna_strt is not None:
            for col_, strt_fill_na in self.dict_fillna_strt.items():
                if col_ not in list_cols:
                    continue
                elif strt_fill_na == "ffill":
                    df_[col_] = df_[col_].ffill()
                elif strt_fill_na == "bfill":
                    df_[col_] = df_[col_].bfill()
                else:
                    raise Exception(f"Invalid fillna strategy: {strt_fill_na}")
        if self.list_cols_dt is not None:
            df_[self.list_cols_dt] = df_[self.list_cols_dt].fillna(self.str_dt_fillna)
        df_[list_cols] = df_[list_cols].fillna(self.str_data_fillna)
        return df_

    def replace_separators(self, df_):
        """
        Replace separators in DataFrame's columns that are numbers with dots as decimal separator
            and commas as thousands separator.

        Args:
            df_ (pandas.DataFrame): DataFrame to replace separators

        Returns:
            pd.DataFrame
        """
        list_cols_numerical = [
            k
            for k, v in self.dict_dtypes.items()
            if v in ["int64", "float64", "int32", "float32", int, float, "int", "float"]
        ]
        for col_ in list_cols_numerical:
            if (df_[col_].dtype == "object") and (
                df_[col_]
                .str.contains(r"^\d{1,3}(?:\.\d{3})*(,\d+)?$", regex=True)
                .any()
            ):
                df_[col_] = df_[col_].str.replace(".", "", regex=False)
                df_[col_] = df_[col_].str.replace(",", ".", regex=False)
                df_[col_] = pd.to_numeric(df_[col_], errors="coerce")
        return df_

    def change_dtypes(self, df_: pd.DataFrame) -> pd.DataFrame:
        dict_dtypes = {
            k: ("str" if v == "date" else v) for k, v in self.dict_dtypes.items()
        }
        df_ = df_.astype(dict_dtypes, errors=self.type_error_action)
        for col_ in self.list_cols_dt:
            if col_ != YAML_GEN["audit_log_cols"]["ref_date"]:
                str_fmt_dt = self.str_fmt_dt
            else:
                str_fmt_dt = YAML_GEN["audit_log_cols"]["str_fmt_dt"]
            df_[col_] = [
                (
                    DatesBR().str_date_to_datetime(d, str_fmt_dt)
                    if str_fmt_dt != "unix_ts"
                    else DatesBR().unix_timestamp_to_date(d, str_tz=self.str_tz)
                )
                for d in df_[col_]
            ]
        list_cols_missing = [
            c
            for c in list(df_.columns)
            if c
            not in ListHandler().extend_lists(
                list(dict_dtypes.keys()), self.list_cols_dt
            )
        ]
        if len(list_cols_missing) > 0:
            df_[list_cols_missing] = df_[list_cols_missing].astype(str)
        return df_

    def strip_hidden_characters(self, df_: pd.DataFrame) -> pd.DataFrame:
        """
        Strips hidden characters (e.g., zero-width spaces) from string columns.

        Args:
            df_ (pd.DataFrame): The DataFrame to clean.

        Returns:
            pd.DataFrame: A DataFrame with hidden characters removed.
        """
        for col_ in [
            c
            for c in df_.select_dtypes(include=["object", "category"]).columns
            if c not in self.list_cols_dt
        ]:
            df_[col_] = df_[col_].str.replace(
                r"[\u200B\u200C\u200D\uFEFF]", "", regex=True
            )
        return df_

    def strip_all_obj_dtypes(self, df_: pd.DataFrame) -> pd.DataFrame:
        list_cols = [
            col_
            for col_ in list(df_.select_dtypes(["object"]).columns)
            if col_ not in self.list_cols_dt
        ]
        for col_ in list_cols:
            try:
                df_[col_] = [x.strip() for x in df_[col_]]
            except AttributeError:
                pass
        return df_

    def remove_duplicated_cols(self, df_: pd.DataFrame) -> pd.DataFrame:
        df_ = df_.loc[:, ~df_.columns.duplicated()]
        return df_

    def data_remove_dupl(self, df_: pd.DataFrame) -> pd.DataFrame:
        df_ = df_.drop_duplicates(
            subset=self.list_cols_drop_dupl, keep=self.strategy_keep_when_dupl
        )
        return df_

    def pipeline(self, df_: pd.DataFrame) -> pd.DataFrame:
        steps = [
            self.check_if_empty,
            self.clean_encoding_issues,
            self.column_names,
            self.limit_columns_to_dtypes,
            self.delete_empty_rows,
            self.completeness,
            self.replace_separators,
            self.change_dtypes,
            self.strip_hidden_characters,
            self.strip_all_obj_dtypes,
            self.remove_duplicated_cols,
            self.data_remove_dupl,
        ]
        df_ = generic_pipeline(df_, steps)
        return df_


class DFStandardizationML(DFStandardization):

    def handle_outliers(self, df_: pd.DataFrame, method="iqr") -> pd.DataFrame:
        """
        Remove or adjust outliers using IQR or Z-score method

        Args:
            method (str): Method to use. Options: ['iqr', 'zscore']

        Returns:
            pd.DataFrame
        """
        for col_ in df_.select_dtypes(include=["number"]).columns:
            if method == "iqr":
                q1, q3 = df_[col_].quantile([0.25, 0.75])
                iqr = q3 - q1
                lower_bound, upper_bound = q1 - 1.5 * iqr, q3 + 1.5 * iqr
                df_[col_] = np.clip(df_[col_], lower_bound, upper_bound)
            elif method == "zscore":
                mean, std = df_[col_].mean(), df_[col_].std()
                df_[col_] = np.clip(df_[col_], mean - 3 * std, mean + 3 * std)
        return df_

    def scale_numeric_data(self, df_: pd.DataFrame, method="minmax") -> pd.DataFrame:
        """
        Normalize or scale numerical data

        Args:
            method (str): Method to use. Options: ['minmax', 'standard']

        Returns:
            pd.DataFrame
        """
        scaler = MinMaxScaler() if method == "minmax" else StandardScaler()
        num_cols = df_.select_dtypes(include=["number"]).columns
        df_[num_cols] = scaler.fit_transform(df_[num_cols])

    def pipeline_ml(
        self,
        df_: pd.DataFrame,
        cols_from_case: Optional[str] = None,
        cols_to_case: Optional[str] = None,
        list_cols_drop_dupl: List[str] = None,
        str_fmt_dt: str = "YYYY-MM-DD",
        method_handle_outliers: str = "iqr",
        method_scale_numeric_data: str = "minmax",
    ) -> pd.DataFrame:
        """
        Pipepline for dataframe cleaning

        Args:
            df_ (pandas.DataFrame): DataFrame to clean
            dict_dtypes (dict): Dictionary of columns and dtypes
            list_cols_dt (list): List of columns to change
            str_dt_fillna (str): Date fillna value
            str_data_fillna (str): Data fillna value
            str_fmt_dt (str): Date format

        Returns:
            pd.DataFrame
        """
        steps = [
            self.pipeline(
                df_, cols_from_case, cols_to_case, list_cols_drop_dupl, str_fmt_dt
            ),
            self.handle_outliers(df_, method_handle_outliers),
            self.scale_numeric_data(df_, method_scale_numeric_data),
        ]
        return generic_pipeline(df_, steps)
